# -*- coding: UTF-8 -*-
import unittest
from mock import patch, Mock
import json
from pandasticsearch.dataframe import DataFrame, Column
from pandasticsearch.types import Greater, BooleanCond


@patch('pandasticsearch.client.urllib.request.urlopen')
def create_ps(mock_urlopen):
    response = Mock()
    dic = {"index": {"mappings": {"type": {"properties": {"a": {"type": "integer"}}}}}}
    response.read.return_value = json.dumps(dic).encode("utf-8")
    mock_urlopen.return_value = response
    return DataFrame.from_es("http://localhost:9200", 'xxx')


class TestDataframe(unittest.TestCase):
    def test_getitem(self):
        ps = create_ps()
        self.assertTrue(isinstance(ps['a'], Column))

        expr = ps['a'] > 2
        self.assertTrue(isinstance(expr, BooleanCond))
        self.assertTrue(isinstance(ps[expr], DataFrame))
        self.assertEqual(ps[expr]._filter, {'range': {'a': {'gt': 2}}})

    def test_columns(self):
        ps = create_ps()
        self.assertEqual(ps.columns, ['a'])

    def test_filter(self):
        ps = create_ps()

        self.assertEqual((ps.filter(ps['a'] > 2))._filter, {'range': {'a': {'gt': 2}}})
        self.assertEqual(ps.where(Greater('a', 2))._filter, {'range': {'a': {'gt': 2}}})


if __name__ == '__main__':
    unittest.main()



# -*- coding: UTF-8 -*-
import unittest

from pandasticsearch.types import *


class TestSchema(unittest.TestCase):
    def test_row(self):
        row = Row(a=1, b='你好,世界')
        print(repr(row))

        self.assertEqual(row['a'], 1)
        self.assertEqual(row['b'], '你好,世界')
        self.assertEqual(repr(row), '''Row(a=1,b='你好,世界')''')

    def test_leaves(self):
        self.assertEqual(GreaterEqual('a', 2).build(), {"range": {"a": {"gte": 2}}})
        self.assertEqual(LessEqual('a', 2).build(), {"range": {"a": {"lte": 2}}})
        self.assertEqual(Less('a', 2).build(), {"range": {"a": {"lt": 2}}})
        self.assertEqual(Equal('a', 2).build(), {"term": {"a": 2}})
        exp = Equal('a', 2)
        self.assertEqual((~exp).build()['bool'], {"must_not": {"term": {"a": 2}}})
        self.assertEqual(Greater('a', 2).build(), {"range": {"a": {"gt": 2}}})
        self.assertEqual(IsIn('a', [1, 2, 3]).build(), {'terms': {'a': [1, 2, 3]}})

    def test_and(self):
        self.assertEqual(
            (GreaterEqual('a', 2) & Less('b', 3)).build()['bool'],
            {
                'must': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'lt': 3}}}]
            })

        self.assertEqual(
            (GreaterEqual('a', 2) & Less('b', 3) & Equal('c', 4)).build()['bool'],
            {
                'must': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'lt': 3}}},
                    {'term': {'c': 4}}]
            })

        self.assertEqual(
            (GreaterEqual('a', 2) & (Less('b', 3) & Equal('c', 4))).build()['bool'],
            {
                'must': [
                    {'range': {'b': {'lt': 3}}},
                    {'term': {'c': 4}},
                    {'range': {'a': {'gte': 2}}}]
            })

    def test_or(self):
        self.assertEqual(
            (GreaterEqual('a', 2) | Less('b', 3)).build()['bool'],
            {
                'should': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'lt': 3}}}]
            })

        self.assertEqual(
            (GreaterEqual('a', 2) | Less('b', 3) | Equal('c', 4)).build()['bool'],
            {
                'should': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'lt': 3}}},
                    {'term': {'c': 4}}]
            })

        self.assertEqual(
            (GreaterEqual('a', 2) | (Less('b', 3) | Equal('c', 4))).build()['bool'],
            {
                'should': [
                    {'range': {'b': {'lt': 3}}},
                    {'term': {'c': 4}},
                    {'range': {'a': {'gte': 2}}}]
            })

    def test_not(self):
        self.assertEqual(
            (~GreaterEqual('a', 2)).build()['bool'],
            {
                'must_not':
                    {'range': {'a': {'gte': 2}}}})

    def test_not_not(self):
        exp = GreaterEqual('a', 2)
        print((~exp).build())
        print((~~exp).build())
        self.assertEqual(
            (~~exp).build()['bool'],
            {
                'must_not': {
                    'bool': {
                        'must_not':
                            {'range': {'a': {'gte': 2}}}}}})

    def test_not_and(self):
        exp = GreaterEqual('a', 2) & Less('b', 3)
        self.assertEqual(
            (~exp).build()['bool'],
            {
                'must_not': {
                    'bool': {
                        'must': [
                            {'range': {'a': {'gte': 2}}},
                            {'range': {'b': {'lt': 3}}}]
                    }
                }
            })

    def test_and_or(self):
        exp = Less('b', 3) | Equal('c', 4)
        actual = GreaterEqual('a', 2) & exp

        self.assertEqual(
            actual.build()['bool'],
            {
                'must': [
                    {'range': {'a': {'gte': 2}}},
                    {
                        'bool': {
                            'should': [
                                {'range': {'b': {'lt': 3}}},
                                {'term': {'c': 4}}
                            ]
                        }
                    }]
            })


if __name__ == '__main__':
    unittest.main()



class Aggregator(object):
    def __init__(self, field):
        self._field = field
        self._aggregator = None

    def build(self):
        return self._aggregator


class MetricAggregator(Aggregator):
    def _metric_agg(self, agg_type, metric_rename=None, params=None):
        if metric_rename is None:
            metric_rename = '{0}({1})'.format(agg_type, self._field)

        agg_field = dict()
        agg_field['field'] = self._field
        if params is not None:
            agg_field.update(params)

        return {metric_rename: {agg_type: agg_field}}


class Avg(MetricAggregator):
    def __init__(self, field):
        super(Avg, self).__init__(field)
        self._aggregator = self._metric_agg('avg')


class Min(MetricAggregator):
    def __init__(self, field):
        super(Min, self).__init__(field)
        self._aggregator = self._metric_agg('min')


class Max(MetricAggregator):
    def __init__(self, field):
        super(Max, self).__init__(field)
        self._aggregator = self._metric_agg('max')


class Cardinality(MetricAggregator):
    def __init__(self, field):
        super(Cardinality, self).__init__(field)
        self._aggregator = self._metric_agg('cardinality')


class ValueCount(MetricAggregator):
    def __init__(self, field):
        super(ValueCount, self).__init__(field)
        self._aggregator = self._metric_agg('value_count')


class CountStar(MetricAggregator):
    def __init__(self):
        super(MetricAggregator, self).__init__('_index')
        self._aggregator = self._metric_agg('value_count', metric_rename='count(*)')


class Percentiles(MetricAggregator):
    def __init__(self, field, percents=None):
        super(Percentiles, self).__init__(field)
        self._aggregator = self._metric_agg('percentiles', params={'percents': percents})


class PercentileRanks(MetricAggregator):
    def __init__(self, field, values=None):
        super(PercentileRanks, self).__init__(field)
        self._aggregator = self._metric_agg('percentile_ranks', params={'values': values})



import json


class Column(object):
    def __init__(self, dim):
        self._dim = dim

    def __eq__(self, other):
        return Equal(dim=self._dim, value=other)

    def __ne__(self, other):
        return ~Equal(dim=self._dim, value=other)

    def __gt__(self, other):
        return Greater(dim=self._dim, value=other)

    def __lt__(self, other):
        return Less(dim=self._dim, value=other)

    def __ge__(self, other):
        return GreaterEqual(dim=self._dim, value=other)

    def __le__(self, other):
        return LessEqual(dim=self._dim, value=other)

    def isin(self, other):
        return IsIn(dim=self._dim, value=other)


class Row(tuple):
    """
    The builtin L{DataFrame} row type for accessing before converted into Pandas DataFrame.
    The fields will be sorted by names.

    >>> row = Row(name="Alice", age=12)
    >>> row
    Row(age=12, name='Alice')
    >>> row['name'], row['age']
    ('Alice', 12)
    >>> row.name, row.age
    ('Alice', 12)
    >>> 'name' in row
    True
    >>> 'wrong_key' in row
    """

    def __new__(cls, **kwargs):
        names = sorted(kwargs.keys())
        row = tuple.__new__(cls, [kwargs[n] for n in names])
        row._fields = names
        return row

    def __getitem__(self, name):
        try:
            idx = self._fields.index(name)
            return super(Row, self).__getitem__(idx)
        except IndexError:
            raise KeyError(name)
        except ValueError:
            raise ValueError(name)

    def __contains__(self, name):
        return name in self._fields

    def __repr__(self):
        return 'Row({0})'.format(','.join(['{0}={1}'.format(k, repr(v)) for k, v in zip(self._fields, tuple(self))]))


# Es filter builder for BooleanCond
class BooleanCond(object):
    def __init__(self, *args):
        self._filter = None

    def __and__(self, x):
        # Combine results
        if isinstance(self, And):
            self.subtree['must'].append(x.subtree)
            return self
        elif isinstance(x, And):
            x.subtree['must'].append(self.subtree)
            return x
        return And(self, x)

    def __or__(self, x):
        # Combine results
        if isinstance(self, Or):
            self.subtree['should'].append(x.subtree)
            return self
        elif isinstance(x, Or):
            x.subtree['should'].append(self.subtree)
            return x
        return Or(self, x)

    def __invert__(self):
        return Not(self)

    @property
    def subtree(self):
        if 'bool' in self._filter:
            return self._filter['bool']
        else:
            return self._filter

    def build(self):
        return self._filter

    def debug_string(self, indent=4):
        return json.dumps(self._filter, indent=indent)


# Binary operator
class And(BooleanCond):
    def __init__(self, *args):
        [isinstance(x, BooleanCond) for x in args]
        super(And, self).__init__()
        self._filter = {'bool': {'must': [x.build() for x in args]}}


class Or(BooleanCond):
    def __init__(self, *args):
        [isinstance(x, BooleanCond) for x in args]
        super(Or, self).__init__()
        self._filter = {'bool': {'should': [x.build() for x in args]}}


class Not(BooleanCond):
    def __init__(self, x):
        assert isinstance(x, BooleanCond)
        super(Not, self).__init__()
        self._filter = {'bool': {'must_not': x.build()}}


# Leaves
class GreaterEqual(BooleanCond):
    def __init__(self, dim, value):
        super(GreaterEqual, self).__init__()
        self._filter = {'range': {dim: {'gte': value}}}


class Greater(BooleanCond):
    def __init__(self, dim, value):
        super(Greater, self).__init__()
        self._filter = {'range': {dim: {'gt': value}}}


class LessEqual(BooleanCond):
    def __init__(self, dim, value):
        super(LessEqual, self).__init__()
        self._filter = {'range': {dim: {'lte': value}}}


class Less(BooleanCond):
    def __init__(self, dim, value):
        super(Less, self).__init__()
        self._filter = {'range': {dim: {'lt': value}}}


class Equal(BooleanCond):
    def __init__(self, dim, value):
        super(Equal, self).__init__()
        self._filter = {'term': {dim: value}}


class IsIn(BooleanCond):
    def __init__(self, dim, value):
        super(IsIn, self).__init__()
        assert isinstance(value, list)
        self._filter = {'terms': {dim: value}}



from __future__ import absolute_import

from pandasticsearch.dataframe import DataFrame, Column
from pandasticsearch.client import RestClient
from pandasticsearch.aggregators import Avg, Max, Min, CountStar, PercentileRanks, ValueCount, Percentiles, Cardinality
from pandasticsearch.queries import Select, Agg
from pandasticsearch.types import Row

col = Column



from pandasticsearch.client import RestClient
from pandasticsearch.queries import Agg, Select
from pandasticsearch.aggregators import Aggregator, CountStar
from pandasticsearch.types import Column, BooleanCond
from pandasticsearch.types import Row

import json
import six
import sys


class DataFrame(object):
    """
    A :class:`DataFrame` treats index and documents in Elasticsearch as named columns and rows.

    >>> from pandasticsearch import DataFrame
    >>> df = DataFrame('http://localhost:9200', index='company')
    """

    def __init__(self, client, columns, schema, **kwargs):
        self._client = client
        self._columns = columns
        self._schema = schema
        self._filter = kwargs.get('filter', None)
        self._aggregation = kwargs.get('aggregation', None)
        self._projection = kwargs.get('projection', None)
        self._limit = kwargs.get('limit', None)
        self._last_query = None

    @staticmethod
    def from_es(url, index, type=None):
        # setup the columns(properties)
        if type is None:
            mapping_endpoint = index
        else:
            mapping_endpoint = index + '/_mapping/' + type

        schema = RestClient(url, mapping_endpoint).get()

        cols = []
        for _, mappings in six.iteritems(schema):
            for _, properties in six.iteritems(mappings['mappings']):
                for k, _ in six.iteritems(properties['properties']):
                    cols.append(k)

        if len(cols) == 0:
            raise Exception('0 columns found in [{0}]'.format(index))

        if type is None:
            endpoint = index + '/_search'
        else:
            endpoint = index + '/' + type + '/_search'
        return DataFrame(RestClient(url, endpoint), cols, schema)

    def __getattr__(self, name):
        """Returns the :class:`Column` denoted by ``name``.
        >>> df.select(df.age).collect()
        [Row(age=2), Row(age=5)]
        """
        if name not in self.columns:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (self.__class__.__name__, name))
        return Column(name)

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            if item not in self._columns:
                raise TypeError('Column does not exist: [{0}]'.format(item))
            return Column(item)
        elif isinstance(item, BooleanCond):
            self._filter = item.build()
            return self
        else:
            raise TypeError('Unsupported expr: [{0}]'.format(item))

    def filter(self, condition):
        """
        Filters rows using a given condition

        where() is an alias for filter().

        :param condition: BooleanCond object
        """
        assert isinstance(condition, BooleanCond)
        return DataFrame(self._client, self._columns, self._schema,
                         filter=condition.build(),
                         aggregation=self._aggregation,
                         projection=self._projection,
                         limit=self._limit)

    where = filter

    def select(self, *cols):
        """
        Projects a set of columns and returns a new L{DataFrame}

        :param cols: list of column names or L{Column}.
        """
        project = {"includes": cols, "excludes": []}
        return DataFrame(self._client, cols, self._schema,
                         filter=self._filter,
                         aggregation=self._aggregation,
                         projection=project,
                         limit=self._limit)

    def limit(self, num):
        """
        Limits the result count to the number specified.
        """
        assert isinstance(num, int)
        assert num >= 1
        return DataFrame(self._client, self._columns, self._schema,
                         filter=self._filter,
                         aggregation=self._aggregation,
                         projection=self._projection,
                         limit=num)

    def agg(self, *aggs):
        """
        Aggregate on the entire DataFrame without groups.
        :param aggs: aggregate functions
        """
        aggregation = {}
        for agg in aggs:
            assert isinstance(agg, Aggregator)
            aggregation.update(agg.build())
        return DataFrame(self._client, self._columns, self._schema,
                         filter=self._filter,
                         aggregation=aggregation,
                         projection=self._projection,
                         limit=self._limit)

    def collect(self):
        """
        Returns all the records as a list of Row.
        >>> df.collect()
        [Row(age=2, name='Alice'), Row(age=5, name='Bob')]

        :return: list of L{Row}
        """
        res_dict = self._client.post(data=self._build_query())
        if self._aggregation is not None:
            query = Agg.from_dict(res_dict)
        else:
            query = Select.from_dict(res_dict)

        return [Row(**v) for v in query.result]

    def to_pandas(self):
        """
        Export to a Pandas DataFrame object.
        :return: The DataFrame representing the query result
        """
        res_dict = self._client.post(data=self._build_query())
        if self._aggregation is not None:
            query = Agg.from_dict(res_dict)
        else:
            query = Select.from_dict(res_dict)
        return query.to_pandas()

    def count(self):
        """
        Returns the number of rows in this index/type.
        >>> df.count()
        2
        """
        return self.agg(CountStar())[0]['count(*)']

    def show(self, n=10):
        """
        Prints the first ``n`` rows to the console.

        :param n:  Number of rows to show.
        """
        assert n > 0
        res_dict = self._client.post(data=self._build_query())
        if self._aggregation is not None:
            query = Agg.from_dict(res_dict)
        else:
            query = Select.from_dict(res_dict)

        cols = self._columns
        widths = []
        tavnit = '|'
        separator = '+'

        for col in cols:
            maxlen = len(col)
            for kv in query.result[:n]:
                if col in kv:
                    s = str(kv[col])
                else:
                    s = '(NULL)'
                if len(s) > maxlen: maxlen = len(s)
            widths.append(min(maxlen, 15))

        for w in widths:
            tavnit += " %-" + "%ss |" % (w,)
            separator += '-' * w + '--+'

        sys.stdout.write(separator + '\n')
        sys.stdout.write(tavnit % tuple(cols) + '\n')
        sys.stdout.write(separator + '\n')
        for kv in query.result[:n]:
            row = []
            for col in cols:
                if col in kv:
                    row.append(str(kv[col]))
                else:
                    row.append('(NULL)')
            sys.stdout.write(tavnit % tuple(row) + '\n')
        sys.stdout.write(separator + '\n')

    def __repr__(self):
        return "DataFrame[%s]" % (", ".join("%s" % c for c in self._columns))

    def print_debug(self):
        """
        Return a indented JSON string returned by the Elasticsearch Server
        """
        sys.stdout.write(json.dumps(self._build_query(), indent=4))

    def print_schema(self):
        """
        Prints out the schema in the tree format.
        >>> df.print_schema()
        index_name
        |-- type_name
          |-- experience :  {'type': 'integer'}
          |-- id :  {'type': 'string'}
          |-- mobile :  {'index': 'not_analyzed', 'type': 'string'}
          |-- regions :  {'index': 'not_analyzed', 'type': 'string'}
        """
        for index, mappings in six.iteritems(self._schema):
            sys.stdout.write('{0}\n'.format(index))
            for typ, properties in six.iteritems(mappings['mappings']):
                sys.stdout.write('|--{0}\n'.format(typ))
                for k, v in six.iteritems(properties['properties']):
                    sys.stdout.write('  |--{0}: {1}\n'.format(k, v))

    @property
    def columns(self):
        """
        Returns all column names as a list.
        :return: column names as a list
        >>> df.columns
        ['age', 'name']
        """
        return self._columns

    @property
    def schema(self):
        return self._schema

    def _build_query(self):
        query = {}

        if self._limit:
            query['size'] = self._limit
        else:
            query['size'] = 20

        if self._aggregation:
            query['aggregations'] = self._aggregation
            query['size'] = 0

        if self._filter:
            query['query'] = {'filtered': {'filter': self._filter}}

        if self._projection:
            query['_source'] = self._projection

        self._last_query = query
        return query



# -*- coding: UTF-8 -*-
import unittest

from pandasticsearch.aggregators import *


class TestAggregators(unittest.TestCase):
    def test_avg(self):
        self.assertEqual(Avg('x').build(), {'avg(x)': {'avg': {'field': 'x'}}})

    def test_max(self):
        self.assertEqual(Max('x').build(), {'max(x)': {'max': {'field': 'x'}}})

    def test_count_star(self):
        self.assertEqual(CountStar().build(), {'count(*)': {'value_count': {'field': '_index'}}})


if __name__ == '__main__':
    unittest.main()
